﻿/*	FT232 7セグ4桁制御ライブラリ
	『昼夜逆転』工作室	http://jsdiy.web.fc2.com

【更新履歴】
2011/11/20	ver 1.00	初版
*/

/*	＊＊＊＊＊ APIリファレンス ＊＊＊＊＊

【構成】

namespace	FT232_7Seg4Lib
{
	struct	SegmentBit	7セグのセグメント名を定義した構造体
	class	PortMgr		回路の配線を示すクラス
	class	SegCtrl		制御処理本体のクラス
}
*/
/*	＊＊＊＊＊ struct SegmentBit ＊＊＊＊＊

【定数】

int	A, B, C, D , E, F, G, DP
int	Null
	7セグのセグメントを個別に指す定数。OR演算で任意の字形を作ることができる。
	SegmentBit.Nullは空白文字（セグメントA～DPが全て消灯している状態）を表す。
*/
/*	＊＊＊＊＊ class PortMgr ＊＊＊＊＊

【コンストラクタ】

PortMgr()
	実際の回路基板上の7セグ、FT232モジュール、シフトレジスタの配線関係（ピンアサイン）を、
	本ライブラリに示すためのクラス。

【メソッド】

//シフトレジスタの出力ポートと7セグのセグメントの対応を設定する
void SetSRegPort(SRegPin pN, SegLedPin segX)
	シフトレジスタの出力ピンP1～P8(P12)が、7セグのセグメントA～DPのどれにつながっているかを示す。

//シフトレジスタの制御ポートとFT232のポートの対応を設定する
void SetSRegCtrl(FtPin data, FtPin clk, FtPin stb)
	シフトレジスタの制御ピンDATA,CLK,STBが、FT232モジュールのI/OピンD0～D7のどれにつながっているかを示す。
	CLRピンは使用しないこととする。

//7セグの桁（コモン）とFT232のポートの対応を設定する
void SetSegLedDigit(FtPin dig1, FtPin dig2, FtPin dig3, FtPin dig4)
	4桁組7セグのコモンピン1～4番が、FT232モジュールのI/OピンD0～D7のどれにつながっているかを示す。
*/
/*	＊＊＊＊＊ class SegCtrl ＊＊＊＊＊

【コンストラクタ】

SegCtrl()
	デフォルトのピンアサインでインスタンス化される。
	つまりデフォルトのピンアサイン通りに実際の基板を製作した場合にのみ、このコンストラクタは有効である。

SegCtrl(PortMgr pm)
	あらかじめ設定したピンアサインを示してインスタンス化する。

【プロパティ】

bool	IsError	 {get}	FT232操作で何らかのエラーが発生した場合true
string	EMessage {get}	エラーメッセージ
int		ECode	 {get}	エラーコード（アプリの終了コード）
	これらプロパティはユーザープログラム側でエラー検出に利用する。
	必要なければ利用しなくてもよい。

【メソッド】

//全桁を消去する
void Cls()
//指定した桁を消去する
void Cls(int keta)
	桁指定はインデックス0から。

//破棄
//・原則としてアプリ終了時にのみ呼び出す。
void Dispose()

//ユーザープログラム側でFTDIオブジェクトを参照したい場合
//・基本的には使用しないメソッド。
FTDI GetFtDevice()

//セグメントパターンの値を取得する：数字
int GetNumberPattern(int num)
	引数numは1桁の数値。
	あらかじめ定義されている、数字'0'～'9'のセグメントパターンを取得する。
	無効な引数の場合、空白文字が返る。

//セグメントパターンの値を取得する：アルファベット
int GetAlphabetPattern(char cc)
	引数ccはアルファベット1文字。
	あらかじめ定義されている、英字'A'～'Z'および'a'～'z'のセグメントパターンを取得する。
	大文字が定義されていないものは小文字を取得する。
	小文字が定義されていないものは大文字を取得する。
	無効な引数の場合、空白文字が返る。

//指定した桁に1文字を表示する
//・アルファベット、空白文字、が対象。
void PutChar(char cc, int keta)
	桁指定はインデックス0から。

//指定した桁から数値を表示する
//・例：(789, 2)->[ ][ ][7][8]	(-2.56, 0)->[-][2.][5][6]	(.34, 1)->[ ][0.][3][4]
void PutNumber(double num, int keta) { PutNumber(num, keta, false); }
void PutNumber(double num, int keta, bool isOverwrite)
	引数numは整数、実数どちらでも対応。
	桁指定はインデックス0から。
	上書き指定をtrueにすると、現在の表示を消去せずに表示する。
	例：[1][2][3][4]に[ ][5][6][ ]を上書きすると[1][5][6][4]と表示される。
	（isOverwrite=falseなら引数の値そのままに[ ][5][6][ ]と表示される）
	ヒント：毎回同じ桁数で同じ位置に表示するならば、isOverwrite=trueとした方が、ちらつきの可能性が抑えられるのでよい。

//指定した桁に1桁の数値を表示する
void PutNumber(int one, int keta)
	0～9の表示専用メソッド。マイナス記号、小数点は不可。
	桁指定はインデックス0から。

//指定した桁にDPを表示する／消去する
void PutSegDP(bool dpOn, int keta)
	現在の表示内容に影響を与えずDPを表示する／消去する。
	桁指定はインデックス0から。

//指定した桁に任意のセグメントパターンを表示する
//・ユーザープログラム側で定義した自由なセグメントパターンを表示することができる。
void PutSegPtn(byte segPtn, int keta)
	桁指定はインデックス0から。
	セグメントを個別に指す定数SegA～SegG,SegDPを組み合わせた任意の図形を表示する。
	例：segPtn = Segment.SegA | Segment.SegD | Segment.SegG ならば[三]の図形が表示される。

//指定した桁から文字列を表示する
void PutString(string str, int keta) { PutString(str, keta, false); }
void PutString(string str, int keta, bool isOverwrite)
	桁指定はインデックス0から。
	引数strが桁数を超える長さの文字列の場合、超える分は無視される（エラーにはならない）。
	上書き指定をtrueにすると、現在の表示を消去せずに表示する。
	例：[a][b][c][d]に[E][ ][ ][F]を上書きすると[E][b][c][F]と表示される。
	（isOverwrite=falseなら引数の値そのままに[E][ ][ ][F]と表示される）
	ヒント：毎回同じ文字数で同じ位置に表示するならば、isOverwrite=trueとした方が、ちらつきの可能性が抑えられるのでよい。

// 7セグ制御開始
//・SegCtrl内のいかなるpublicメソッドにも先だって実行すること。
bool Start() { return Start(0, false); }
bool Start(uint devidx) { return Start(devidx, false); }
bool Start(bool errReturn) { return Start(0, errReturn); }
bool Start(uint devidx, bool errReturn)
	devidx…FT232Rモジュールを指すインデックス。
		PCに接続したFT232Rモジュールが1個だけならdevidx=0が確定しているので特に引数で渡す必要はない。
	errReturn…trueにすると、デバイス・オープンエラーのとき、リトライせずに呼び出し元へ戻る。
		falseのとき（引数に指定しないとき）は最大10回までリトライを試みる。
	
	FT232RモジュールはPCに接続した直後、自動的に接続確認動作を開始し、約5秒で完了する。
	その間にアプリケーションを実行するとデバイス・オープンエラーとなる。
	このとき、errReturn=trueであればリトライせずに呼び出し元に制御を戻すことができ、
	「Wait...」など画面に表示することができる。
	その後errReturn=falseとして（または引数なしで）再度Start()を呼び出し、リトライから処理を続行させればよい。
	もし接続待ちの表示が不要であれば、最初からerrReturn=falseとして（または引数なしで）1回だけStart()を呼び出す。
	この場合、デバイス・オープンエラーが発生した後、続けて1秒おきに最大10回リトライされる。

//ユーザープログラム側からステータスをリセットする
//・エラーを無視してプログラムを続行させたい場合に呼び出す。
//・エラー情報のプロパティがリセットされるだけであり、エラー状態が回復するわけではない。
void ErrorStatusReset()
	プロパティ IsError, EMessage, ECodeをリセットする。
*/

using System;
using System.Threading;		//Thread, Sleep
using FTD2XX_NET;

namespace FT232_7Seg4Lib
{
	/*	秋月電子FT232RLモジュールのピン配置
	   +----------+
	D0 |TXD    GND|
	D4 |DTR#   CB0|
	D2 |RTS#   CB1|
	   |VIO    VCC|
	D1 |RXD    RST|
	D7 |RI#    3V3|
	   |GND    CB3|
	D5 |DSR#   PU1|
	D6 |DCD#   PU2|
	D3 |CTS#   VCC|
	   |CB4    USB|(5V)
	   |CB2    SLD|(GND)
	   +----↓----+
	   USBコネクタ側
	*/
	/*	シフトレジスタNJU3711のピン配置
		 +----v----+
		 |P3    VDD|(5V)
		 |P4     P2|
		 |P5     P1|
	(GND)|VSS  CLR#|
		 |P6   STB#|
		 |P7    CLK|
		 |P8   DATA|
		 +---------+
	*/
	/*	シフトレジスタNJU3714のピン配置
		 +----v----+
		 |P5    VDD|(5V)
		 |P6     P4|
		 |P7     P3|
		 |P8     P2|
	(GND)|VSS    P1|
		 |P9    VSS|(GND)
		 |P10  CLR#|
		 |P11  STB#|
		 |P12   CLK|
		 |SO   DATA|
		 +---------+
	*/
	/*	【7セグのセグメント図】
	+-A-+
	F   B
	+-G-+
	E   C
	+-D-+ DP
	*/
	
	/*	7セグ／シフトレジスタ／FT232RLモジュールのピンのデフォルト対応
	7セグのセグメント：					A	B	C	D	E	F	G	DP	d1	d2	d3	d4
	シフトレジスタ	 ：	DATA  CLK  STB	P1	P2	P3	P4	P5	P6	P7	P8
	FT232Rモジュール ：	D0	  D1   D2									D3	D4	D5	D6
	*/
	
	//FT232のビットバンモード信号名とインデックス
	enum FtPin { D0 = 0, D1, D2, D3, D4, D5, D6, D7 };
	
	//SRegのパラレル出力ポート名とインデックス
	//・実際の部品としてNJU7311(8bit)またはNJU7314(12bit)を使うものとした定義。
	//・名前の数字とインデックス値の対応に注意。
	enum SRegPin { P1 = 0, P2, P3, P4, P5, P6, P7, P8, P9, P10, P11, P12 };
	
	//7セグのセグメント名とインデックス
	enum SegLedPin { A = 0, B, C, D, E, F, G, DP };
	
	//シフトレジスタのピンアサインを示す
	//・PortMgrの内部で使用する。
	struct PortShiftReg
	{
		public int Data { get; private set; }
		public int Clk { get; private set; }
		public int StbRdy { get; private set; }
		public int StbGo { get { return 0; } }	//負論理の0（定数扱い）
		//未使用	public int ClrRdy { get; private set; }
		//未使用	public int ClrGo { get { return 0; } }	//負論理の0（定数扱い）
		
		//シフトレジスタの制御ポートとFT232のポートの対応を設定する（ビットマスク）
		public void SetSRegCtrl(FtPin data, FtPin clk, FtPin stb)
		{
			Data = 1 << (int)data;
			Clk = 1 << (int)clk;
			StbRdy = 1 << (int)stb;
		}
	}
	
	//7セグのピンアサインを示す
	//・PortMgrの内部で使用する。
	struct PortSegLed
	{
		public int[] Dig { get; private set; }
		
		//7セグの桁（コモン）とFT232のポートの対応を設定する（ビットマスク）
		public void SetSegLedDigit(FtPin dig1, FtPin dig2, FtPin dig3, FtPin dig4)
		{
			Dig = new int[] { 1 << (int)dig1, 1 << (int)dig2, 1 << (int)dig3, 1 << (int)dig4 };
		}
	}
	
	//デバイスのピンアサインを示す
	class PortMgr
	{
		private static PortShiftReg devSReg;
		private static PortSegLed devSegLed;
		
		public static PortShiftReg SReg { get { return devSReg; } }
		public static PortSegLed SegLed { get { return devSegLed; } }
		
		//シフトレジスタのポートPnの先が7セグのセグメントA-G,DPのどれにつながっているかを示す配列
		private static SegLedPin[] pinSRegToSegs;
		
		//コンストラクタ
		public PortMgr()
		{
			devSReg.SetSRegCtrl(FtPin.D0, FtPin.D1, FtPin.D2);
			devSegLed.SetSegLedDigit(FtPin.D3, FtPin.D4, FtPin.D5, FtPin.D6);
			
			pinSRegToSegs = new SegLedPin[SegCtrl.SHIFTREG_USEBIT];
			for (int idxPn = 0, idxSegPin = 0; idxPn < pinSRegToSegs.Length; idxPn++, idxSegPin++)
				pinSRegToSegs[idxPn] = (SegLedPin)idxSegPin;
		}

		//シフトレジスタの出力ポートと7セグのセグメントの対応を設定する
		public void SetSRegPort(SRegPin pN, SegLedPin segX)
		{
			pinSRegToSegs[(int)pN] = segX;
		}

		//シフトレジスタの制御ポートとFT232のポートの対応を設定する
		//・ユーザープログラム側へ中継するために用意された関数
		public void SetSRegCtrl(FtPin data, FtPin clk, FtPin stb)
		{
			devSReg.SetSRegCtrl(data, clk, stb);
		}

		//7セグの桁（コモン）とFT232のポートの対応を設定する
		//・ユーザープログラム側へ中継するために用意された関数
		public void SetSegLedDigit(FtPin dig1, FtPin dig2, FtPin dig3, FtPin dig4)
		{
			devSegLed.SetSegLedDigit(dig1, dig2, dig3, dig4);
		}
		
		//セグメントA-DPの順番をシフトレジスタのポートP1-Pnの順に並べ替える
		//・srcBitAry[]にはセグメントA-DPの順にビット0/1が格納されている（字形を表している）。
		//・pinSRegToSegs[]にはP1-Pnの順に、それに対応するセグメント(A-DP)が格納されている。
		public static void SetOrderedBitArray(int[] srcBitAry, int[] distBitArray)
		{
			//srcBitAry[],distBitArray[],pinSRegToSegs[]の長さは全て一致しているのが正しい状態だが、
			//ユーザプログラム側からの設定ミスを考慮してtry-catchしておく。
			try
			{
				for (int idxPn = 0, idxSegPin; idxPn < pinSRegToSegs.Length; idxPn++)
				{
					idxSegPin = (int)pinSRegToSegs[idxPn];
					distBitArray[idxPn] = srcBitAry[idxSegPin];
				}
			}
			catch
			{
				try { Array.Copy(srcBitAry, distBitArray, Math.Min(srcBitAry.Length, distBitArray.Length)); }
				catch { }
			}
		}
	}
	
	//7セグ1桁が持つ値
	class SegKeta
	{
		private int segPattern;		//セグメントパターン（SegmentPatternの定義値そのもの）
		public int SegPattern
		{
			get { return segPattern; }
			set
			{
				segPattern = value;
				SetBitArray(segPattern, BitArray);
				PortMgr.SetOrderedBitArray(BitArray, OrderedBitArray);
			}
		}

		public int[] BitArray { get; private set; }			//セグメントパターン定義通りの並び順
		public int[] OrderedBitArray { get; private set; }	//シフトレジスタのポート(P0-Pn)順にセグメント(SegX)を並べ替えたもの	※Write()用

		//コンストラクタ
		public SegKeta(int numOfSegment)
		{
			if (numOfSegment < 0) numOfSegment = 0;
			BitArray = new int[numOfSegment];
			OrderedBitArray = new int[numOfSegment];
		}

		//セグメントパターンの2進数表現をビット配列に変換する
		private static void SetBitArray(int srcVal, int[] distBitArray)
		{
			//srcValはMSB側がセグメントDP、LSBがセグメントAである
			//distBitArray[0]=セグメントA、distBitArray[last]=セグメントDPとなるよう格納する
			try
			{
				for (int i = 0; i < distBitArray.Length; i++, srcVal >>= 1)
					distBitArray[i] = srcVal & 0x01;
			}
			catch
			{
				for (int i = 0; i < distBitArray.Length; i++)
					distBitArray[i] = 0;
			}
		}
	}
	
	//セグメントのビット定義
	//・enumだと論理演算でいちいちキャストしなければならないのでstructで定義する。
	struct SegmentBit
	{
		public const int	//PGFEDCBA
			A	= 0x01,		//00000001
			B	= 0x02,		//00000010
			C	= 0x04,		//00000100
			D	= 0x08,		//00001000
			E	= 0x10,		//00010000
			F	= 0x20,		//00100000
			G	= 0x40,		//01000000
			DP	= 0x80,		//10000000
			Null	= 0x00;	//00000000

		/*	16セグの定義例	//                        21   21  21
		 *					//00000000_0000000P_MLKJIHGG_FEDDCBAA	(32bit)
		 *	A1	= 0x0001,
		 *	A2	= 0x0002,
		 *	B	= 0x0004,
		 *	C	= 0x0008,
		 *	D1	= 0x0010,
		 *	D2	= 0x0020,
		 *	E	= 0x0040,
		 *	…
		 *	M	= 0x8000,
		 *	DP	= 0x10000,
		 *	A	= A1 | A2,
		 *	D	= D1 | D2,
		 *	G	= G1 | G2,
		 *	Null	= 0x0000;
		 *	*/
	}
	
	//セグメントパターン（字形）の定義
	class SegmentPattern
	{
		//数字の字形
		private static readonly string[] PtnNumber =
#region
		{
		//	"0GFEDCBA"
			"00111111",	//0
			"00000110",	//1
			"01011011",	//2
			"01001111",	//3
			"01100110",	//4
			"01101101",	//5
			"01111101",	//6
			"00100111",	//7
			"01111111",	//8
			"01101111"	//9
		};
#endregion
		//アルファベット大文字の字形
		private static readonly string[] PtnUpperCase =
#region
		{
		//	"0GFEDCBA"
			"01110111",	//A
			"00000000",	//(B)
			"00111001",	//C
			"00000000",	//(D)
			"01111001",	//E
			"01110001",	//F
			"00111101",	//G
			"01110110",	//H
			"00110000",	//I
			"00011110",	//J
			"01110101",	//K
			"00111000",	//L
			"00110111",	//M（'U'の逆さ）
			"00000000",	//(N)
			"00111111",	//O（'0'と同じ）
			"01110011",	//P
			"00111011",	//Q
			"00000000",	//(R)
			"01101101",	//S（'5'と同じ）
			"00110001",	//T（'「'の形）
			"00111110",	//U
			"01110010",	//V
			"01111110",	//W（'A'の逆さ）
			"01100100",	//X
			"01101110",	//Y
			"00011011"	//Z
		};
#endregion
		//アルファベット小文字の字形
		private static readonly string[] PtnLowerCase =
#region
		{
		//	"0GFEDCBA
			"01011111",	//a
			"01111100",	//b
			"01011000",	//c
			"01011110",	//d
			"01111011",	//e
			"00000000",	//(f)
			"00000000",	//(g)
			"01110100",	//h
			"00010000",	//i
			"00001110",	//j
			"00000000",	//(k)
			"00110000",	//l（'I'と同じ）
			"00000000",	//(m)
			"01010100",	//n
			"01011100",	//o
			"00000000",	//(p)
			"01100111",	//q
			"01010000",	//r
			"00000000",	//(s)
			"01111000",	//t
			"00011100",	//u
			"00000000",	//(v)
			"00000000",	//(w)
			"00000000",	//(x)
			"00000000",	//(y)
			"00000000"	//(z)
		};
#endregion
		
		//ヌル文字の字形
		private const string BITS_NULL = "00000000";

		//数字のセグメントパターン値を取得する
		public int GetNumber(int num)
		{
			string bits = (0 <= num && num <= 9) ? PtnNumber[num] : BITS_NULL;
			int segPtnValue = BitStringToValue(bits);
			return segPtnValue;
		}

		//アルファベットのセグメントパターン値を取得する
		public int GetAlphabet(char cc)
		{
			string bits = BITS_NULL;

			if ('A' <= cc && cc <= 'Z')
			{
				int idx = cc - 'A';
				bits = PtnUpperCase[idx];
				if (bits == BITS_NULL) bits = PtnLowerCase[idx];
			}
			else if ('a' <= cc && cc <= 'z')
			{
				int idx = cc - 'a';
				bits = PtnLowerCase[idx];
				if (bits == BITS_NULL) bits = PtnUpperCase[idx];
			}

			int segPtnValue = BitStringToValue(bits);
			return segPtnValue;
		}

		//2進数を表す文字列をint値に変換する
		private static int BitStringToValue(string bits)
		{
			int n = 0;
			try
			{
				for (int i = 0; i < bits.Length; i++)
					n = (n << 1) | (bits[i] == '0' ? 0 : 1);
			}
			catch { n = 0; }
			return n;
		}
	}
	
	//エラー情報
	struct Status
	{
		public enum ECode { NormalTermination = 0, FT_OpenError = 10, FT_InitializeError, FT_Error, OtherReason = 255 };
		public bool isError { get; set; }
		public string message { get; private set; }
		public ECode ecode { get; private set; }
		
		public void Reset() { SetMes(null, ECode.NormalTermination, false); }
		public void SetMes(string mes, ECode code) { SetMes(mes, code, true); }
		public void SetMes(string mes, ECode code, bool err)
		{
			message = mes;
			ecode = code;
			isError = err;
		}
	}
	
	//制御処理本体
	class SegCtrl
	{
		public const int SHIFTREG_USEBIT = 8;	//シフトレジスタを何ビット使用するか（何ビットのシフトレジスタ部品を使うかではない）
		public const int NUMOF_SEGMENT = 8;		//SegA-G,DPでセグメントが8個
		public const int NUMOF_KETA = 4;		//7セグの桁数（1個何桁の部品かではなく、全体で何桁制御するか）
		
		private Thread thdd;
		private static FTDI ftdev;
		private static SegKeta[] leds;
		private static PortMgr portMgr;
		private SegmentPattern segmentPattern = new SegmentPattern();

		private static Status status;	//スレッドからも参照するのでstatic
		//ユーザープログラム側から参照するプロパティ
		public bool IsError { get { return status.isError; } set { status.isError = value; } }
		public string EMessage { get { return status.message; } }
		public int ECode { get { return (int)status.ecode; } }


		//コンストラクタ
		public SegCtrl(PortMgr pm)
		{
			portMgr = (pm == null) ? new PortMgr() : pm;
			
			ftdev = null;
			thdd = null;

			leds = new SegKeta[NUMOF_KETA];
			for (int i = 0; i < leds.Length; i++)
				leds[i] = new SegKeta(NUMOF_SEGMENT);

			status.Reset();
		}
		//
		public SegCtrl()
			: this((PortMgr)null)	//分かりやすさのため明示的にキャストする
		{ }

		//ユーザープログラム側からステータスをリセットする
		//・エラーを無視してプログラムを続行させたい場合に呼び出す。
		//・エラー情報のプロパティがリセットされるだけであり、エラー状態が回復するわけではない。
		public void ErrorStatusReset()
		{
			status.Reset();
		}

		// 7セグ制御開始
		//・SegCtrl内のいかなるpublicメソッドにも先だって実行すること。
		public bool Start() { return Start(0, false); }
		public bool Start(uint devidx) { return Start(devidx, false); }
		public bool Start(bool errReturn) { return Start(0, errReturn); }
		public bool Start(uint devidx, bool errReturn)
		{
			//Start()成功後の二重呼び出しをチェックする
			if (ftdev != null && ftdev.IsOpen && thdd != null && thdd.IsAlive) return true;

			//FTDIオブジェクトを生成する
			if (ftdev == null) ftdev = new FTDI();
			
			for (int retryCount = 0; retryCount < 10; retryCount++)
			{
				if (ftdev.OpenByIndex(devidx) == FTDI.FT_STATUS.FT_OK)
				{
					if (ftdev.SetBitMode(0xFF, FTDI.FT_BIT_MODES.FT_BIT_MODE_ASYNC_BITBANG) == FTDI.FT_STATUS.FT_OK
					//	&& ftdev.SetBaudRate(9600 * 2) == FTDI.FT_STATUS.FT_OK	//任意でコメントマークを外してもよい
						)
					{
						//ダイナミックドライブ開始
						thdd = new Thread(ThDinamicDrive);
						//thdd.Priority = ThreadPriority.Highest;	//任意でコメントマークを外してもよい
						thdd.Start();
						return true;
					}
					status.SetMes("FT232 initialize error.", Status.ECode.FT_InitializeError, false);
					return false;
				}
				else if (errReturn)
				{
					break;
				}
				Thread.Sleep(1000);
			}
			status.SetMes("FT232 open error.", Status.ECode.FT_OpenError, false);
			return false;
		}
		
		//ユーザープログラム側でFTDIオブジェクトを参照したい場合
		//・基本的には使用しないメソッド。
		public FTDI GetFtDevice()
		{
			return ftdev;
		}
		
		//破棄
		//・原則としてアプリ終了時にのみ呼び出す。
		public void Dispose()
		{
			//不自然なゴミが表示されないよう、最初に表示を消去する
			//・シフトレジスタの全出力ポートをLoにする処理も兼ねている。
			if (ftdev != null && ftdev.IsOpen && thdd != null && thdd.IsAlive)
			{
				Cls();
			}
			
			if (thdd != null)
			{
				//スレッド破棄
				ReqThreadStop = true;
				thdd.Join();
				thdd = null;
			}

			if (ftdev != null)
			{
				if (ftdev.IsOpen)
				{
					//FT232 全出力ポートをLoにする
					//・7セグの桁制御ポートとシフトレジスタの制御ポートがLoになる。
					uint writtenLength = 0;
					if (ftdev.Write(new byte[] { 0x00 }, 1, ref writtenLength) != FTDI.FT_STATUS.FT_OK)
						status.SetMes("FT232 write error.", Status.ECode.FT_Error);

					//FT232 送受信バッファクリア
					if (ftdev.Purge(FTDI.FT_PURGE.FT_PURGE_RX | FTDI.FT_PURGE.FT_PURGE_TX) != FTDI.FT_STATUS.FT_OK)
						status.SetMes("FT232 purge error.", Status.ECode.FT_Error);
				}
				ftdev.Close();
				ftdev = null;
			}
		}

		//【スレッド】ダイナミックドライブ
		private static volatile bool ReqThreadStop;
		private static void ThDinamicDrive()
		{
			if (ftdev == null || !ftdev.IsOpen) return;	//保険
			ReqThreadStop = false;
			int keta = 0, prevKeta;
			byte[] wbuf;
			
			while (!ReqThreadStop)
			{
				//次の桁
				prevKeta = keta;
				if (++keta == leds.Length) keta = 0;

				//セグメントパターンをセットするデータ配列を作る
				wbuf = GetTxData(leds[keta].OrderedBitArray, keta, prevKeta);
				
				//データを出力する
				Write(wbuf);
			}
		}

		//シフトレジスタへ入力するデータを設定する
		//・現在の桁の点灯を維持しつつ、次の桁のデータを設定する。
		private static byte[] StTxBuf = new byte[SHIFTREG_USEBIT * 2 + 1];
		private static byte[] GetTxData(int[] dataBits, int nextKeta, int currentKeta)
		{
			//dataBits[]にはP1,P2,..,Pnの順にデータが格納されている。
			//シフトレジスタのDATAへはPn,..,P2,P1の順で入力することに注意する。
			for (int tidx = 0, i = SHIFTREG_USEBIT - 1; 0 <= i; i--, tidx += 2)
			{
				StTxBuf[tidx] = (byte)(PortMgr.SegLed.Dig[currentKeta] | PortMgr.SReg.Data * dataBits[i] | PortMgr.SReg.StbRdy);
				StTxBuf[tidx + 1] = (byte)(StTxBuf[tidx] | PortMgr.SReg.Clk);
			}
			StTxBuf[StTxBuf.Length - 1] = (byte)(PortMgr.SegLed.Dig[nextKeta] | PortMgr.SReg.StbGo);

			return StTxBuf;
		}
		
		//FT232 送信バッファにbyte配列を書き込む
		private static void Write(byte[] datas)
		{
			uint writtenLength = 0;
			if (ftdev.Write(datas, datas.Length, ref writtenLength) != FTDI.FT_STATUS.FT_OK)
				status.SetMes("FT232 write error.", Status.ECode.FT_Error);
		}
		
		//ユーザープログラム側からセグメントパターンの値を取得する：数字
		public int GetNumberPattern(int num)
		{
			return segmentPattern.GetNumber(num);
		}
		
		//ユーザープログラム側からセグメントパターンの値を取得する：アルファベット
		public int GetAlphabetPattern(char cc)
		{
			return segmentPattern.GetAlphabet(cc);
		}
		
		//桁インデックスの妥当性をチェックする
		//・桁指定のあるpublicなメソッドから呼び出す。
		private bool KetaValidation(int keta)
		{
			return (0 <= keta && keta < NUMOF_KETA);
		}

		//指定した桁に任意のセグメントパターンを表示する
		//・ユーザープログラム側で定義した自由なセグメントパターンを表示することができる。
		public void PutSegPtn(byte segPtn, int keta)
		{
			if (!KetaValidation(keta)) return;
			leds[keta].SegPattern = segPtn;
		}

		//指定した桁にDPを表示する／消去する
		public void PutSegDP(bool dpOn, int keta)
		{
			if (!KetaValidation(keta)) return;
			if (dpOn)
				leds[keta].SegPattern |= SegmentBit.DP;
			else
				leds[keta].SegPattern &= ~SegmentBit.DP;
		}
		
		//指定した桁を消去する
		public void Cls(int keta)
		{
			if (!KetaValidation(keta)) return;
			leds[keta].SegPattern = SegmentBit.Null;
		}

		//全桁を消去する
		//・ダイナミックドライブは動作しているのでFT232Rモジュールが「全ポートLo」の状態になるわけではない。
		public void Cls()
		{
			for (int i = 0; i < NUMOF_KETA; i++) Cls(i);
		}

		//指定した桁に1文字を表示する
		//・アルファベット、定義済みの記号、空白文字、が対象。
		public void PutChar(char cc, int keta)
		{
			PutString(cc.ToString(), keta, true);
		}

		//指定した桁に1桁の数値を表示する
		public void PutNumber(int one, int keta)
		{
			if (!KetaValidation(keta)) return;
			leds[keta].SegPattern = segmentPattern.GetNumber(one);
		}
		
		//指定した桁から数値を表示する
		//・例：(789, 2)->[ ][ ][7][8]	(-2.56, 0)->[-][2.][5][6]	(.34, 1)->[ ][0.][3][4]
		public void PutNumber(double num, int keta) { PutNumber(num, keta, false); }
		public void PutNumber(double num, int keta, bool isOverwrite)
		{
			PutString(num.ToString(), keta, isOverwrite);
		}
		
		//指定した桁から文字列を表示する
		public void PutString(string str, int keta) { PutString(str, keta, false); }
		public void PutString(string str, int keta, bool isOverwrite)
		{
			if (!KetaValidation(keta)) return;
			if (string.IsNullOrEmpty(str)) return;
			if (!isOverwrite) Cls();

			int startKeta = keta;
			char prevChar = '\0';	//「.」以外なら何でもよい
			foreach (char cc in str)
			{
				if (NUMOF_KETA == keta && cc != '.') break;
				if (NUMOF_KETA < keta) break;
				
				if ('0' <= cc && cc <= '9')
				{
					leds[keta].SegPattern = segmentPattern.GetNumber(cc - '0');
				}
				else if (cc == ' ')
				{
					leds[keta].SegPattern = SegmentBit.Null;
				}
				else if (cc == '-')
				{
					leds[keta].SegPattern = SegmentBit.G;
				}
				else if (cc == '.')
				{
					if (startKeta < keta && prevChar != '.') keta--;
					leds[keta].SegPattern |= SegmentBit.DP;
				}
				else
				{
					leds[keta].SegPattern = segmentPattern.GetAlphabet(cc);
				}
				
				prevChar = cc;
				keta++;
			}
		}
		
	}	//<- end of class

}	//<- end of namespace
