﻿//	『昼夜逆転』工作室	http://jsdiy.web.fc2.com
//
//	FT232RLモジュール 7セグ制御 サンプルアプリ
//	「セグメントアタック」
//
//	2011/11/10	v1.00	初版

/*	【作者コメント】
	Gameクラス／Start()メソッドの記述内容は、
	マイコン用のCのソースをほぼそのままコピー＆ペーストし、少し手直ししただけのものです。
	当時、意図的に関数を分けずswitch()内に詰め込む記述をしていました。
	今回移植するにあたり、敢えてそのままの形にしてあります。
	プログラミング初心者は決してソース記述の手本にしないでください。
*/

using System;
using System.Timers;	//Timer
using FT232_7Seg4Lib;

namespace ft232r7seg4satk
{
	enum GStatus
	{
		ST_TITLE,
		ST_INIT,
		ST_PLAY,
		ST_HIT,
		ST_OVER,
		ST_SCORE,
		ST_HISCORE
	};
	
	class Game
	{
		private const int
			PLAYER_LIFETIME_DEFAULT = (1500 / 250),	//1.5秒（250ms単位）
			PLAYER_LIFETIME_FASTEST = (1000 / 250), //1.0秒（250ms単位）
			ENEMY_MOVEWAIT_DEFAULT = (PLAYER_LIFETIME_DEFAULT * 7),
			SCORE_MAX = 999;
		
		private readonly SegCtrl segled;
		private readonly Timer timer;
		private static bool tci;
		
		//コンストラクタ
		public Game(SegCtrl segled)
		{
			this.segled = segled;
			tci = false;

			timer = new Timer(250);		//移植元プログラムの割り込み時間が250msだったのでそれに合わせる
			timer.Elapsed += new ElapsedEventHandler(OnTimeElapsed);
		}

		//タイマー時間経過
		private static void OnTimeElapsed(object source, ElapsedEventArgs e)
		{
			tci = true;
		}

		//乱数生成
		private static readonly Random RandomGenerator = new Random();
		private int rand() { return RandomGenerator.Next(); }

		//ゲーム実行本体
		public void Start()
		{
			bool[] swOnA = new bool[2];		//キー入力状態
			int	player = SegmentBit.Null, playerLifeTime = 0, prevPlayer = SegmentBit.Null;
			int	enemy = SegmentBit.Null, enemyPos = 0, enemyMoveWait = 0;
			bool isCreatePlayer = true, isCreateEnemy = true;
			int	hitWaitCnt = 0;
			int	hitCount = 0;
			int	nowPlayerLifetime = 0, nowEnemyMoveWait = 0;
			int	score = 0, hiscore = 0;
			GStatus state = GStatus.ST_TITLE;
			bool gameoverEffectSegDpOn = false;

			//タイマー開始
			tci = false;
			timer.Start();

			//タイトル表示
			segled.PutString("S.ATK", 0);

			//メインループ
			while (true)
			{
				//キー入力判定
				swOnA[0] = swOnA[1] = false;
				if (Console.KeyAvailable)
				{
					ConsoleKeyInfo keyInfo = Console.ReadKey(true);
					while (Console.KeyAvailable) Console.ReadKey(true);		//キーバッファクリア（連打対策）
					if (keyInfo.Key == ConsoleKey.Spacebar)
					{
						if (keyInfo.Modifiers == ConsoleModifiers.Shift) swOnA[1] = true;
						else swOnA[0] = true;
					}
					if (keyInfo.Key == ConsoleKey.Escape) break;	//メインループを抜ける
				}

				//動作状態
				switch (state)
				{
				case GStatus.ST_INIT:
					segled.Cls();
					player = SegmentBit.Null;
					isCreatePlayer = true;
					isCreateEnemy = true;
					hitCount = 0;
					nowPlayerLifetime = PLAYER_LIFETIME_DEFAULT;
					nowEnemyMoveWait = ENEMY_MOVEWAIT_DEFAULT;
					score = 0;
					state = GStatus.ST_PLAY;
					break;

				case GStatus.ST_PLAY:
					//プレイヤーの弾を生成する
					if (isCreatePlayer)
					{
						isCreatePlayer = false;
						do { player = 1 << (rand() % 7); } while (prevPlayer == player);	//前回の形と同じにならないようにする
						prevPlayer = player;	//今回生成した形を保存しておく
						playerLifeTime = 0;
						segled.PutSegPtn((byte)player, 0);
					}

					//敵を生成する
					if (isCreateEnemy)
					{
						isCreateEnemy = false;
						enemy = SegmentBit.Null;
						do
						{
							for (int i = 0; i < 7; i++)
								enemy |= ((rand() % 100 < 30 ? 1 : 0) << i);
						}
						while (enemy == SegmentBit.Null);	//空白の敵にならないようにする
						enemyPos = 3;	//敵の出現位置（4桁目）
						enemyMoveWait = 0;
						segled.PutSegPtn((byte)enemy, enemyPos);
					}

					//発射
					if (swOnA[0])
					{
						//敵にセグメントをはめる、取る
						if ((enemy & player) != 0) enemy &= ~player;
						else enemy |= player;
						segled.PutSegPtn((byte)enemy, enemyPos);

						//この弾を無効にする（表示を消す）
						player = SegmentBit.Null;
						segled.PutSegPtn((byte)player, 0);

						//数字形成を判定する
						bool isMatch = false;
						for (int i = 0; i < 10 && !isMatch; i++)
							if (enemy == segled.GetNumberPattern(i)) isMatch = true;
						if (isMatch)	//敵を倒した
						{
							//スコア計算
							int i, bits, bitCnt;
							for (bits = enemy, bitCnt = 0, i = 0; i < 7; i++, bits >>= 1)
								bitCnt += (bits & 1);	//セグメントの個数をカウントする
							score += bitCnt + enemyPos;
							if (SCORE_MAX < score) score = SCORE_MAX;

							//スピード調節
							if (++hitCount % 10 == 0)
							{
								if (PLAYER_LIFETIME_FASTEST < nowPlayerLifetime)
								{
									nowPlayerLifetime--;
									nowEnemyMoveWait = nowPlayerLifetime * 7;
								}
							}

							segled.PutSegDP(true, enemyPos);
							hitWaitCnt = 0;
							state = GStatus.ST_HIT;
							break;
						}

						//数字を作らず、敢えてセグメントを全消ししたらボーナス点獲得
						if (enemy == 0) score += 20;
					}

					if (tci)	//タイマー時間経過
					{
						tci = false;

						//一定時間が経過したらプレイヤーの弾を変更する
						if (++playerLifeTime == nowPlayerLifetime)
							isCreatePlayer = true;

						//一定時間が経過したら敵を移動する
						if (++enemyMoveWait == nowEnemyMoveWait)
						{
							enemyMoveWait = 0;
							segled.PutSegPtn(SegmentBit.Null, enemyPos);
							enemyPos--;
							segled.PutSegPtn((byte)enemy, enemyPos);
						}
						if (enemyPos == 0)	//プレイヤーと衝突
						{
							if (hiscore < score) hiscore = score;
							state = GStatus.ST_OVER;
							break;
						}
					}
					break;

				case GStatus.ST_HIT:
					if (tci)
					{
						tci = false;
						if (++hitWaitCnt == 4)		//250ms単位で1sec待つ（1sec/250ms=4カウント）
						{
							segled.PutSegDP(false, enemyPos);
							segled.PutSegPtn(SegmentBit.Null, enemyPos);

							isCreatePlayer = true;
							isCreateEnemy = true;
							state = GStatus.ST_PLAY;
						}
					}
					break;

				case GStatus.ST_OVER:
					if (tci)
					{
						tci = false;
						gameoverEffectSegDpOn = !gameoverEffectSegDpOn;		//点滅効果
						segled.PutSegDP(gameoverEffectSegDpOn, 0);
					}

					if (swOnA[0])
					{
						setScoreLED(score, 'P');
						state = GStatus.ST_SCORE;
					}
					break;

				case GStatus.ST_SCORE:
					if (swOnA[0])
					{
						setScoreLED(hiscore, 'H');
						state = GStatus.ST_HISCORE;
					}
					break;

				case GStatus.ST_HISCORE:
					if (swOnA[0])
					{
						segled.PutString("S.ATK", 0);
						state = GStatus.ST_TITLE;
					}
					break;

				default:	//ST_TITLE
					if (swOnA[0])
					{
						state = GStatus.ST_INIT;
					}
					if (swOnA[1])
					{
						setScoreLED(hiscore, 'H');
						state = GStatus.ST_HISCORE;
					}
					break;
				}

			}	//<- end of while()

			//タイマー停止／破棄
			timer.Stop();
			timer.Dispose();
		}

		//スコア表示
		private void setScoreLED(int score, char mark)
		{
			//例：(24, 'P') = [P.][ ][2][4]
			string sScore = string.Format("{0}.{1,3:D}", mark, score);
			segled.PutString(sScore, 0);
		}
		
	}	//<- end of class
	
}	//<- end of namespace
